import os, sys

def create8xv( appvarData, appvarName, fileName="", comment="Comment", archived=False ):
#	Creates an appvar with the specified data and name in the current directory.
#	If fileName is specified, that's the name given to the file on the computer.
#	If comment is specified, it's added as a comment in the file (42 bytes max).
#	If archived is True, the default transfer location will be the Flash memory.
	if fileName == "":
		fileName = appvarName + ".8xv"
#	Header data
	fileData = b"**TI83F*\x1A\x0A\x0A"
#	Comment
	fileData += bytes( comment.ljust(42,"\x00")[0:42], "latin-1" )
#	Size + 19
	dataSize = len( appvarData )
	fileData += numToBytes( dataSize + 19 )
#	Version? Will have to look up
	fileData += bytes( [0x0D, 0x00] )
#	Size + 2
	fileData += numToBytes( dataSize + 2 )
#	Variable type (appvar = 0x15)
	fileData += bytes( [0x15] )
#	Name
	fileData += bytes( appvarName.ljust(8,"\x00")[0:8] + "\x00", "latin-1" )
#	Archive status ( 0x00 = RAM, 0x80 = Archive )
	if archived:
		fileData += bytes( [0x80] )
	else:
		fileData += bytes( [0x00] )
#	Size + 2
	fileData += numToBytes( dataSize + 2 )
#	Size
	fileData += numToBytes( dataSize )
#	Actual data
	fileData += appvarData
#	Checksum
	checksum = sum( fileData[55:] ) % 65536
	fileData += numToBytes( checksum )	
#	Create file
	with open( fileName, "wb" ) as f:
		f.write( fileData )

def numToBytes( num ):
#	Converts number to sequence of two little-endian bytes.
	return bytes([num%256,num//256])

def printError( error ):
#	Prints a custom error or something, then exits Python.
	print( "\nError: " + error )
	sys.exit()

def replaceStringIndex( str, index, replace ):
#	Replaces index of string with given replacement string.
	return str[:index]+replace+str[index+1:]

def stringWidth( str ):
#	Returns the "width" of a string in calculator pixels.
	return sum([[0,6,4,4,4,4,4,5,4,4,4,4,4,4,4,5,4,5,4,4,5,5,4,5,6,5,4,4,5,6,3,3,1,2,4,6,6,4,5,2,3,3,6,4,3,4,2,4,4,4,4,4,4,4,4,4,4,4,2,3,4,4,4,4,6,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,3,4,4,3,4,4,4,4,4,3,4,4,2,4,4,3,6,4,4,4,4,4,3,3,4,4,6,4,4,5,4,2,4,5,4,4,3,4,4,4,4,4,4,4,4,5,5,5,5,5,5,5,5,5,5,4,4,4,4,4,4,4,4,4,4,4,4,4,4,6,6,6,6,6,6,6,6,4,4,4,4,5,5,5,5,4,4,5,5,3,3,4,4,2,5,4,5,6,4,4,3,4,5,6,5,5,5,5,6,6,4,4,4,3,4,3,4,3,4,4,4,0,4,4,4,4,4,3,3,4,3,3,6,5,6,5,6,5,6,5,6,5,6,5,2,3,5,6,4,6,5,5,4,4,6,0,0,0,0,0,0,0,0][ord(c)] for c in str])

#	ERROR CHECK: File to read from must be specified and exist.
try:
	open( sys.argv[1], "r" )
except IndexError:
#	If no file specified, explain how to use the editor.
	print( "\n[ Puzzler's Crossword Editor ]\n" )
	print( "Converts .txt files to crossword appvars to use with Puzzler's Crossword.\n" )
	print( "  python crossword.py [file] [output]\n" )
	print( "[file] is a text file formatted as such:" )
	print( "  Crossword solution data in ASCII format." )
	print( "    (Use # for black squares, uppercase letters for letters in solution, and" )
	print( "    lowercase letters for highlighted/circled letters in solution." )
	print( "  One blank line." )
	print( "  The word ACROSS, followed by the ACROSS clues as they would be listed in a" )
	print( "  newspaper." )
	print( "  One blank line." )
	print( "  The word DOWN, followed by the DOWN clues as they would be listed in a" )
	print( "  newspaper." )
	print( "An example file, CROSSWRD.txt, is available as a reference. Keep in mind" )
	print( "that one-letter words are not recognized and will not receive clues.\n" )
	print( "[output] is an optional name and path for the outputted .8xv file." )
	sys.exit()
except FileNotFoundError:
	printError( "File does not exist!" )

#	fileName contains the name of the file passed to the second argument.
#	fileNameNoExt gets rid of its extension.
fileName = os.path.basename(sys.argv[1])
fileNameNoExt = os.path.splitext(fileName)[0]

try:
	targetName = os.path.basename(sys.argv[2])
except IndexError:
	targetName = fileNameNoExt + ".8xv"

with open( sys.argv[1], "r" ) as f:
	print( "Found file: " + fileName + "\nReading crossword from file..." )
	
	firstLine = True
	line = f.readline().rstrip()
	while line != "":
		if firstLine:
			firstLine = False
			crossword = ["#" * ( len(line) + 2 )]
		crossword += ["#" + line + "#"]
		line = f.readline().rstrip()
	crossword += [crossword[0]]
#	ERROR CHECK: Each of the rows of the crossword must be of equal length.
	if min( [ len( c ) for c in crossword ] ) is not max( [ len( c ) for c in crossword ] ):
		printError( "Inconsistent row lengths!" )
	crosswordSizeX = len( crossword[0] ) - 2
	crosswordSizeY = len( crossword ) - 2
#	ERROR CHECK: Both crossword dimensions must be larger than 1.
	if crosswordSizeX <= 1 or crosswordSizeY <= 1:
		printError( "Crossword too small!" )
	
	print( "Reading clues from file..." )
	
#	ERROR CHECK: The first line of ACROSS clues should read "ACROSS".
	if f.readline().rstrip() != "ACROSS":
		printError( "ACROSS clues must be labeled and placed before DOWN clues!" )
	acrossClues = []
	line = f.readline().rstrip()
	while line != "":
		acrossClues += [line]
		line = f.readline().rstrip()
#	ERROR CHECK: The first line of DOWN clues should read "DOWN".
	if f.readline().rstrip() != "DOWN":
		printError( "DOWN clues must be labeled and placed after ACROSS clues!" )
	downClues = []
	line = f.readline().rstrip()
	while line != "":
		downClues += [line]
		line = f.readline().rstrip()

print( "Processing crossword data..." )

crosswordData = [ crosswordSizeX, crosswordSizeY ]
for row in range( 1, len( crossword ) - 1 ):
	for col in range( 1, len( crossword[0] ) - 1 ):
		c = crossword[row][col]
#		ERROR CHECK: Crossword data must entirely consist of letters, "#", and " ".
#		Yes, I know I said nothing about " " in the usage notes,
#		but it's an empty space, and why would you ever need it?
		try:
			addedData = " ABCDEFGHIJKLMNOPQRSTUVWXYZ#".index( c.upper() )
		except ValueError:
			printError( "Invalid character in crossword!" )
		if c.islower():
			addedData |= 0b00100000
		if c != "#":
			if ( crossword[row-1][col] == "#" ):
				if ( crossword[row+1][col] != "#" ):
					addedData |= 0b01000000
			if ( crossword[row][col-1] == "#" ):
				if ( crossword[row][col+1] != "#" ):
					addedData |= 0b10000000
		crosswordData += [addedData]

print( "Processing clue data..." )

clueGrid = [[0 for i in range(crosswordSizeX)] for j in range(crosswordSizeY)]
acrossClueGrid = [[0 for i in range(crosswordSizeX)] for j in range(crosswordSizeY)]
downClueGrid = [[0 for i in range(crosswordSizeX)] for j in range(crosswordSizeY)]
clue = acrossClue = downClue = 0
for row in range( crosswordSizeY ):
	for col in range( crosswordSizeX ):
		c = crosswordData[ row * crosswordSizeX + col + 2 ]
		if c & 0b11000000:
			clue += 1
			clueGrid[row][col] = clue
		if c & 0b10000000:
			acrossClue += 1
			acrossClueGrid[row][col] = acrossClue
		if c & 0b01000000:
			downClue += 1
			downClueGrid[row][col] = downClue
#	ERROR CHECK: There must be at least enough ACROSS clues for all ACROSS clue spaces.
if len( acrossClues ) < max( [ max( a ) for a in acrossClueGrid ] ):
	printError( "Not enough ACROSS clues!" )
acrossClueData = ""
for row in range( crosswordSizeY ):
	for col in range( crosswordSizeX ):
		if acrossClueGrid[row][col]:
			addAcrossClueData = str( clueGrid[row][col] ) + " ACROSS: " + acrossClues[ acrossClueGrid[row][col]-1 ] + " "
			firstWord = True
			i = j = 0
			while i < len( addAcrossClueData ):
				j = addAcrossClueData.find( " ", j, len( addAcrossClueData ) )
				if stringWidth( addAcrossClueData[i:j] ) < 96:
					j = j + 1
					firstWord = False
				else:
					if firstWord == True:
						j = i
						while stringWidth( addAcrossClueData[i:j] ) < 96:
							j += 1
						i = j - 1
						addAcrossClueData = replaceStringIndex( addAcrossClueData, i-1, addAcrossClueData[i-1] + "\xFF" )
					else:
						i = addAcrossClueData.rfind( " ", 0, j-1 ) + 1
						j = i
						addAcrossClueData = replaceStringIndex( addAcrossClueData, i-1, "\xFF" )
						firstWord = True
				if j >= len( addAcrossClueData ) - 1:
					i = len( addAcrossClueData )
					addAcrossClueData = replaceStringIndex( addAcrossClueData, i-1, "\x00" )
			acrossClueData += addAcrossClueData
#	ERROR CHECK: There must be at least enough DOWN clues for all DOWN clue spaces.
if len( downClues ) < max( [ max( d ) for d in downClueGrid ] ):
	printError( "Not enough DOWN clues!" )
downClueData = ""
for col in range( crosswordSizeX ):
	for row in range( crosswordSizeY ):
		if downClueGrid[row][col]:
			addDownClueData = str( clueGrid[row][col] ) + " DOWN: " + downClues[ downClueGrid[row][col]-1 ] + " "
			firstWord = True
			i = j = 0
			while i < len( addDownClueData ):
				j = addDownClueData.find( " ", j, len( addDownClueData ) )
				if stringWidth( addDownClueData[i:j] ) < 96:
					j = j + 1
					firstWord = False
				else:
					if firstWord == True:
						j = i
						while stringWidth( addDownClueData[i:j] ) < 96:
							j += 1
						i = j - 1
						addDownClueData = replaceStringIndex( addDownClueData, i-1, addDownClueData[i-1] + "\xFF" )
					else:
						i = addDownClueData.rfind( " ", 0, j-1 ) + 1
						j = i
						addDownClueData = replaceStringIndex( addDownClueData, i-1, "\xFF" )
				if j >= len( addDownClueData ) - 1:
					i = len( addDownClueData )
					addDownClueData = replaceStringIndex( addDownClueData, i-1, "\x00" )
			downClueData += addDownClueData
crosswordData += [ len( acrossClues ) ]
crosswordData += bytes( acrossClueData, "latin-1" ) + bytes( downClueData, "latin-1" )

print( "Creating " + targetName + "..." )

create8xv( bytes( crosswordData ), fileNameNoExt, fileName=targetName, comment="Created by the Puzzler's Crossword Editor", archived=True)

print( "Success!" )